from .basis import BSplineBasis
from .splineobject import SplineObject
from .curve import Curve
from .surface import Surface
from .volume import Volume
from .trimmedsurface import TrimmedSurface
from .splinemodel import SplineModel

__version__ = '1.5.8'
